// ===========================================================================
// CONTENT  : CLASS ObjectIdGenerator
// AUTHOR   : M.Duchrow
// VERSION  : 2.0 - 11/102014
// HISTORY  :
//  22/02/2008  mdu  CREATED
//  11/10/2014  mdu  added  -> implements StringGenerator
//
// Copyright (c) 2008-2014, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pf.text ;

// ===========================================================================
// IMPORTS
// ===========================================================================
import org.pf.bif.identifier.IObjectIdGenerator;
import org.pfsw.text.StringGenerator;

/**
 * Generates numeric IDs that are left padded with zeros.
 *
 * @author M.Duchrow
 * @version 2.0
 */
public class ObjectIdGenerator implements IObjectIdGenerator, StringGenerator
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  protected final static char DEFAULT_PAD_CHAR 			= '0' ;
  protected final static long DEFAULT_START_ID	  	= 1 ;
  protected final static int DEFAULT_LENGTH				  = 10 ;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private int length = DEFAULT_LENGTH ;
  private char padChar = DEFAULT_PAD_CHAR ;
  private String prefix = null ;

  private long nextAvailableId = DEFAULT_START_ID ;
  protected long getNextAvailableId() { return nextAvailableId ; }
  protected void setNextAvailableId( long newValue ) { nextAvailableId = newValue ; }

  protected int bufferLen = DEFAULT_LENGTH ;
  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   * That is an ID length of 10 and a start ID of 1 with fill character '0'.
   */
  public ObjectIdGenerator()
  {
    super();
    this.setLength(this.getDefaultLength());
    this.setNextAvailableId(this.getDefaultStartId());
    this.setPadChar(this.getDefaultPadChar());
  } // ObjectIdGenerator() 

  // -------------------------------------------------------------------------

  /**
   * Initialize the new instance with the length for the generated identifiers.
   * 
   * @param idLength The length to which Ids are filled up with leading zeros (must be > 0)
   */
  public ObjectIdGenerator(int idLength)
  {
    this();
    this.setLength(idLength);
  } // ObjectIdGenerator() 

  // -------------------------------------------------------------------------

  /**
   * Initialize the new instance with the length for the generated identifiers
   * and the id to start with.
   * 
   * @param startId The first id to be generated
   * @param idLength The length to which Ids are filled up with leading zeros
   */
  public ObjectIdGenerator(long startId, int idLength)
  {
    this(idLength);
    this.setNextAvailableId(startId);
  } // ObjectIdGenerator() 

  // -------------------------------------------------------------------------

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns a new unique identifier as string.
   */
  @Override
  public synchronized String newIdentifier()
  {
    return (this.leftPad(this.nextIdentifier()));
  } // newIdentifier() 

  // -------------------------------------------------------------------------

  /**
   * Returns a new identifier which is different to the last one.
   */
  public long nextIdentifier()
  {
    long id = this.getNextId();
    this.setNextId(id + 1);
    return id;
  } // nextIdentifier() 

  // -------------------------------------------------------------------------
  
  /**
   * Returns the full length of the generated identifiers. That includes the 
   * length of the prefix if one is defined.
   * 
   * @see #getPrefix()
   * @see ObjectIdGenerator#getLength()
   */
  public int getFullLength() 
  {
    return this.getPrefixLength() + this.getLength();
  } // getFullLength()
  
  // -------------------------------------------------------------------------

  /** 
   * Returns the length to which the IDs (the numeric part without prefix)) 
   * are filled up (left padded).
   */
  public int getLength()
  {
    return length;
  } // getLength() 

  // -------------------------------------------------------------------------

  /**
   * Set the length to which the IDs (numeric part without prefix) must be 
   * filled-up with the defined padding character.
   * Since all IDs are based on data type long a value greater than 20 is
   * not really reasonable, however it is allowed here. 
   * 
   * @param len The new length (must be > 0).
   * @see #getPadChar()
   * @see #setPadChar(char)
   */
  public void setLength(int len)
  {
    if (len > 0)
    {
      length = len;
      this.calcBufferLength();
    }
  } // setLength() 

  // -------------------------------------------------------------------------

  /**
   * Set the character that is used to fill up the IDs to the full length.
   * 
   * @param fillChar The fill character
   */
  public void setPadChar(char fillChar)
  {
    padChar = fillChar;
  } // setPadChar() 

  // -------------------------------------------------------------------------

  /**
   * Return the character that is used to fill up the IDs to the full length.
   */
  public char getPadChar()
  {
    return padChar;
  } // getPadChar() 

  // -------------------------------------------------------------------------

  /**
   * Returns the prefix that will be prepended to every generated ID.
   */
  public String getPrefix()
  {
    return prefix;
  } // getPrefix() 

  // -------------------------------------------------------------------------

  /**
   * Set the prefix that will be prepended to every generated ID.
   */
  public void setPrefix(String newValue)
  {
    prefix = newValue;
    this.calcBufferLength();
  } // setPrefix() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns whether or not a prefix is defined.
   */
  public boolean hasPrefix() 
  {
    return this.getPrefix() != null;
  } // hasPrefix()
  
  // -------------------------------------------------------------------------

  // -------------------------------------------------------------------------
  // Interface StringGenerator
  // -------------------------------------------------------------------------
  /**
   * Returns a new string. It is a new identifier as returned by {@link #newIdentifier()}.
   * 
   * @return A new string containing a left padded number with the currently defined length, plus the optional prefix.
   */
  @Override
  public String generateString()
  {
    return this.newIdentifier();
  } // generateString()
  
  // -------------------------------------------------------------------------
  
  /**
   * Generates a new string (i.e. identifier) with the specified length. 
   * The contents of the string is a number left padded with the currently
   * defined padding character and prefixed with the defined prefix (if any is set).
   * <p>
   * The specified length will be used only for this single string generation.
   * That is, the defined length (see {@link #getLength()} and {@link #setLength(int)})
   * of this object will not be used or changed by this method.
   * 
   * @param len The length of the string to generate.
   * @return A new string with the given length.
   * @throws IllegalArgumentException if the given length is negative.
   */
  @Override
  public String generateString(int len)
  {
    int savedLength;
    int lengthWithoutPrefix;
    String string;
    
    if (len < 0)
    {
      throw new IllegalArgumentException("The length for a newly generated string must not be negative: " + len);
    }
    savedLength = this.getLength();
    lengthWithoutPrefix = len - this.getPrefixLength();
    if (lengthWithoutPrefix <= 0)
    {
      throw new IllegalArgumentException("The underlying prefix is too long for the given length. No space left for any generated string part. Prefix: " + this.getPrefix());
    }    
    this.setLength(lengthWithoutPrefix);
    string = this.generateString();
    this.setLength(savedLength);
    return string;
  } // generateString()

  // -------------------------------------------------------------------------
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected int getPrefixLength() 
  {
    if (this.hasPrefix())
    {
      return this.getPrefix().length();
    }
    return 0;
  } // getPrefixLength()
  
  // -------------------------------------------------------------------------

  protected String leftPad(long id)
  {
    StringBuffer buffer;

    buffer = new StringBuffer(bufferLen);
    if (this.hasPrefix())
    {
      buffer.append(this.getPrefix());
    }
    this.str().leftPadCh(buffer, id, this.getLength(), this.getPadChar());
    return buffer.toString();
  } // leftPad() 

  // -------------------------------------------------------------------------

  protected long getNextId()
  {
    return this.getNextAvailableId();
  } // getNextId() 

  // -------------------------------------------------------------------------

  protected void setNextId(long id)
  {
    this.setNextAvailableId(id);
  } // setNextId() 

  // -------------------------------------------------------------------------

  protected void calcBufferLength()
  {
    if (prefix == null)
    {
      bufferLen = this.getLength();
    }
    else
    {
      bufferLen = this.getLength() + prefix.length();
    }
  } // calcBufferLength() 

  // -------------------------------------------------------------------------

  protected long getDefaultStartId()
  {
    return DEFAULT_START_ID;
  } // getDefaultStartId() 

  // -------------------------------------------------------------------------

  protected int getDefaultLength()
  {
    return DEFAULT_LENGTH;
  } // getDefaultLength() 

  // -------------------------------------------------------------------------

  protected char getDefaultPadChar()
  {
    return DEFAULT_PAD_CHAR;
  } // getDefaultPadChar() 

  // -------------------------------------------------------------------------

  protected StringUtil str()
  {
    return StringUtil.current();
  } // str() 

  // -------------------------------------------------------------------------

} // class ObjectIdGenerator 
