// ===========================================================================
// CONTENT  : CLASS Base64Converter
// AUTHOR   : M.Duchrow
// VERSION  : 2.0 - 16/09/2014
// HISTORY  :
//  22/05/2004  mdu  CREATED
//	11/06/2005	mdu		changed	-->	Due to changed underlying Base64 class
//	08/03/2009	mdu		bugfix	--> Avoid \n in long results
//  26/04/2014  mdu   added   --> base64url support (see RFC-4648, section 5)
//  16/09/2014  mdu   changed --> added base64url support for byte[] and auto-padding for decode
//
// Copyright (c) 2004-2014, by M.Duchrow. All rights reserved.
// ===========================================================================
package org.pf.util;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.UnsupportedEncodingException;

/**
 * A converter that allows to encode strings, byte arrays and char arrays to
 * BASE64 and vice versa.<p>
 * <b> 
 * Currently it is based on the class Base64 from Robert Harder 
 * (http://iharder.sourceforge.net/base64).
 * Thanks to him that he published his implementation as open source.
 * </b>
 * This class mainly adds some convenience methods for string handling.
 * 
 * @author M.Duchrow
 * @version 2.0
 */
public class Base64Converter
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================

  // =========================================================================
  // PUBLIC CLASS METHODS
  // =========================================================================
  /**
   * Returns a base64url encoded version of the given string.
   */
  public static String encodeBase64url(String unencoded)
  {
    String base64;
    String base64url;
    
    base64 = encode(unencoded);
    base64url = toBase64url(base64);
    return base64url;
  } // encodeBase64url() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns a BASE64 encoded version of the given string
   */
  public static String encode(String unencoded)
  {
    byte[] bytes;

    bytes = unencoded.getBytes();
    return encodeToString(bytes);
  } // encode() 

  // -------------------------------------------------------------------------

  /**
   * Returns a BASE64 encoded version of the given string, where the 
   * specified character encoding is used before the base64 encoding.
   * @throws UnsupportedEncodingException If the given char encoding is not supported.
   */
  public static String encode(String unencoded, String charEncoding) throws UnsupportedEncodingException
  {
    byte[] bytes;

    bytes = unencoded.getBytes(charEncoding);
    return encodeToString(bytes);
  } // encode() 

  // -------------------------------------------------------------------------

  /**
   * Returns a base64url encoded version of the given string, where the 
   * specified character encoding is used before the base64url encoding.
   * @throws UnsupportedEncodingException If the given char encoding is not supported.
   */
  public static String encodeBase64url(String unencoded, String charEncoding) throws UnsupportedEncodingException
  {
    String base64;
    
    base64 = encode(unencoded, charEncoding);
    return toBase64url(base64);
  } // encodeBase64url() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns a BASE64 encoded version of the given string with new lines
   * after all 76 characters (MIME compliant).
   */
  public static String encodeWithLineBreaks(String unencoded)
  {
    byte[] bytes;

    bytes = unencoded.getBytes();
    return encodeToStringWithLineBreaks(bytes);
  } // encodeWithLineBreaks() 

  // -------------------------------------------------------------------------

  /**
   * Returns a BASE64 encoded version of the given character array
   */
  public static char[] encode(char[] unencoded)
  {
    String str;

    str = new String(unencoded);
    return encode(str.getBytes());
  } // encode() 

  // -------------------------------------------------------------------------

  /**
   * Returns a BASE64 encoded version of the given byte array.
   */
  public static char[] encode(byte[] unencoded)
  {
    return encodeToString(unencoded).toCharArray();
  } // encode() 

  // -------------------------------------------------------------------------

  /**
   * Returns a BASE64 encoded version of the given byte array 
   * with new lines after all 76 bytes (MIME compliant).
   */
  public static char[] encodeWithLineBreaks(byte[] unencoded)
  {
    return encodeToStringWithLineBreaks(unencoded).toCharArray();
  } // encodeWithLineBreaks() 

  // -------------------------------------------------------------------------

  /**
   * Returns a BASE64 encoded version of the given byte array as String
   */
  public static String encodeToString(byte[] unencoded)
  {
    return Base64.encodeBytes(unencoded, Base64.DONT_BREAK_LINES);
  } // encodeToString() 

  // -------------------------------------------------------------------------

  /**
   * Returns a BASE64url encoded string version of the given byte array.
   */
  public static String encodeBase64url(byte[] bytes)
  {
    String base64Str;
    
    base64Str = encodeToString(bytes);
    return toBase64url(base64Str);
  } // encodeBase64url() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns a BASE64 encoded version of the given byte array as String
   * with a new line (\n) after all 76 bytes (MIME compliant).
   */
  public static String encodeToStringWithLineBreaks(byte[] unencoded)
  {
    return Base64.encodeBytes(unencoded);
  } // encodeToStringWithLineBreaks() 

  // -------------------------------------------------------------------------

  /**
   * Returns a BASE64 encoded version of the given char array as String
   */
  public static String encodeToString(char[] unencoded)
  {
    char[] chars;

    chars = encode(unencoded);
    return new String(chars);
  } // encodeToString() 

  // -------------------------------------------------------------------------

  /**
   * Returns a byte array decoded from the given BASE64 encoded char array
   */
  public static byte[] decode(char[] encoded)
  {
    return decode(new String(encoded));
  } // decode() 

  // -------------------------------------------------------------------------

  /**
   * Returns a byte array decoded from the given BASE64 encoded String
   */
  public static byte[] decode(String encoded)
  {
    return Base64.decode(autoPadding(encoded));
  } // decode() 

  // -------------------------------------------------------------------------

  /**
   * Returns a decoded byte array from the given base64url encoded String.
   * 
   * @param base64url The base64url encoded string (must not be null).
   */
  public static byte[] decodeFromBase64urlString(String base64url)
  {
    String base64;
    
    base64 = toBase64(base64url);
    return decode(base64);
  } // decodeFromBase64urlString() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns a string decoded from the given base64url encoded String.
   * 
   * @param base64url The base64url encoded string.
   */
  public static String decodeToBase64urlString(String base64url)
  {
    String base64;
    
    base64 = toBase64(base64url);
    return decodeToString(base64);
  } // decodeToBase64urlString() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns a string decoded from the given BASE64 encoded String
   * 
   * @param encoded The BASE64 encoded string
   */
  public static String decodeToString(String encoded)
  {
    byte[] bytes;

    bytes = decode(encoded);
    return new String(bytes);
  } // decodeToString() 

  // -------------------------------------------------------------------------

  /**
   * Returns a string decoded from the given BASE64 encoded String
   * 
   * @param encoded The BASE64 encoded string (must not be null).
   * @param charEncoding The name of the encoded string's character encoding (e.g. "UTF-8"). 
   */
  public static String decodeToString(String encoded, String charEncoding) throws UnsupportedEncodingException
  {
    byte[] bytes;

    bytes = decode(encoded);
    return new String(bytes, charEncoding);
  } // decodeToString() 

  // -------------------------------------------------------------------------

  /**
   * Returns a string decoded from the given base64url encoded String
   * 
   * @param base64url The base64url encoded string (must not be null).
   * @param charEncoding The name of the encoded string's character encoding (e.g. "UTF-8"). 
   */
  public static String decodeToBase64urlString(String base64url, String charEncoding) throws UnsupportedEncodingException
  {
    String base64;
    
    base64 = toBase64(base64url);
    return decodeToString(base64, charEncoding);
  } // decodeToBase64urlString() 
  
  // -------------------------------------------------------------------------
  
  // =========================================================================
  // PRIVATE CLASS METHODS
  // =========================================================================
  private static String toBase64url(final String base64)
  {
    String base64url;
    
    base64url = base64.replace('+', '-');
    base64url = base64url.replace('/', '_');
    return base64url;
  } // toBase64url() 
  
  // -------------------------------------------------------------------------
  
  private static String toBase64(final String base64url)
  {
    String base64;
    
    base64 = base64url.replace('-', '+');
    base64 = base64.replace('_', '/');
    return base64;
  } // toBase64() 
  
  // -------------------------------------------------------------------------
  
  private static String autoPadding(final String string)
  {
    int missing;
    String newString;
    
    newString = string;
    missing = string.length() % 4;
    for (int i = 1; i <= missing; i++)
    {
      newString += "=";
    }
    return newString;
  } // autoPadding()

  // -------------------------------------------------------------------------
  
  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  private Base64Converter()
  {
    super();
  } // Base64Converter() 

} // class Base64Converter 
