// ===========================================================================
// CONTENT  : CLASS CaseInsensitiveKeyMap
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.0 - 15/11/2014
// HISTORY  :
//  04/09/2002  duma  CREATED
//	30/07/2004	duma	added		-->	Initial capacity in constructor
//  15/11/2014  mdu   refactored  --> Not subclassing MapWrapper anymore
//
// Copyright (c) 2002-2014, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pf.util;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * This class provides a map that treats all keys as case-insensitive strings.
 * That means, the methods get(key) and containsKey(key) will find entries
 * regardless of the keys case.
 * Be aware that this is to some degree NOT an exact implementation of the Map
 * interface. Iterators on the keys and values are NOT backed by the 
 * corresponding collection. 
 * That means that particularly a Collections.synchronizedMap() doesn't 
 * work with an instance of this class!
 *
 * @author Manfred Duchrow
 * @version 2.0
 */
public class CaseInsensitiveKeyMap<V> implements Map<String, V>, Serializable
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final long serialVersionUID = 5699414864704470568L;
  // =========================================================================
  
  // INSTANCE VARIABLES
  // =========================================================================
  private Map<String, NamedValue<V>> internalMap;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public CaseInsensitiveKeyMap()
  {
    this(20);
  } // CaseInsensitiveKeyMap() 

  // -------------------------------------------------------------------------

  /**
   * Initialize the new instance with an initial capacity.
   * 
   * @param initialCapacity Defines how many "slots" for elements are created initially
   */
  public CaseInsensitiveKeyMap(int initialCapacity)
  {
    super();
    this.internalMap(this.createInternalMap(initialCapacity));
  } // CaseInsensitiveKeyMap() 

  // -------------------------------------------------------------------------

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  @Override
  public boolean isEmpty()
  {
    return this.internalMap().isEmpty();
  } // isEmpty() 

  // -------------------------------------------------------------------------

  @Override
  public int size()
  {
    return this.internalMap().size();
  } // size() 

  // -------------------------------------------------------------------------

  @Override
  public void clear()
  {
    this.internalMap().clear();
  } // clear()

  // -------------------------------------------------------------------------

  @Override
  public Set<String> keySet()
  {
    return this.internalMap().keySet();
  } // keySet()

  // -------------------------------------------------------------------------

  /**
   * This method returns true, if the given key can be found in this map.
   * 
   * @see java.util.Map#containsKey(Object)
   */
  @Override
  public boolean containsKey(Object key)
  {
    String name;

    name = this.standardizeKey(key);
    if (name == null)
    {
      return false;
    }
    return this.internalMap().containsKey(name);
  } // containsKey() 

  // -------------------------------------------------------------------------

  /**
   * Returns the object that was registered under the specified key.
   * The given key must be a String. If not null is returned.
   * 
   * @see java.util.Map#get(Object)
   */
  @Override
  public V get(Object key)
  {
    V value = null;
    String name;
    NamedValue<V> namedValue;

    name = this.standardizeKey(key);
    if (name != null)
    {
      namedValue = this.internalMap().get(name);
      if (namedValue != null)
      {
        value = namedValue.value();
      }
      return value;
    }
    return null;
  } // get() 

  // -------------------------------------------------------------------------

  /**
   * Puts the given value under the specified key into the map.    <br>
   * <b>If the key is no java.lang.String nothing will be put to the map
   * and the method returns null!</b>
   * 
   * @see java.util.Map#put(Object, Object)
   */
  @Override
  public V put(String key, V value)
  {
    NamedValue<V> namedValue;
    String name;

    name = this.standardizeKey(key);
    if (name != null)
    {
      namedValue = new NamedValue<V>(key, value);
      namedValue = this.internalMap().put(name, namedValue);
      if (namedValue == null)
      {
        return null;
      }
      return namedValue.value();
    }
    return null;
  } // put() 

  // -------------------------------------------------------------------------

  /**
   * Each entry of the given map will be added. 
   * Each added entry gets an associated timestamp.
   * 
   * @see java.util.Map#putAll(Map)
   */
  @Override
  public void putAll(Map<? extends String, ? extends V> map)
  {
    for (Map.Entry<? extends String, ? extends V> entry : map.entrySet())
    {
      this.put(entry.getKey(), entry.getValue());
    }
  } // putAll() 

  // -------------------------------------------------------------------------

  /**
   * This method returns true, if the given value is a value of any entry
   * in this map.
   * 
   * @see java.util.Map#containsValue(Object)
   */
  @Override
  public boolean containsValue(Object value)
  {
    Map.Entry<String, NamedValue<V>> entry;

    entry = this.findByValue(value);
    return (entry != null);
  } // containsValue() 

  // -------------------------------------------------------------------------

  /**
   * Returns all values of the map.
   * 
   * @see java.util.Map#values()
   */
  @Override
  public Collection<V> values()
  {
    Collection<V> values;

    values = new ArrayList<V>(this.size());
    for (NamedValue<V> namedValue : this.internalMap().values())
    {      
      values.add(namedValue.value());
    }
    return values;
  } // values() 

  // -------------------------------------------------------------------------

  /**
   * Removes the entry with the specified key.
   * 
   * @see java.util.Map#remove(Object)
   */
  @Override
  public V remove(Object key)
  {
    String name;
    NamedValue<V> namedValue;

    if (key instanceof String)
    {
      name = this.standardizeKey(key);
      if (name != null)
      {
        namedValue = this.internalMap().remove(name);
        if (namedValue != null)
        {
          return namedValue.value();
        }
      }
    }
    return null;
  } // remove() 

  // -------------------------------------------------------------------------

  /**
   * This method automatically <i>touches</i> every entry, so that its
   * expire period starts anew.
   * 
   * @see java.util.Map#entrySet()
   */
  @Override
  public Set<Map.Entry<String, V>> entrySet()
  {
    Set<Map.Entry<String, V>> entries;
    Map.Entry<String, V> newEntry;
    NamedValue<V> namedValue;

    entries = new HashSet<Map.Entry<String, V>>(this.size());
    for (Map.Entry<String, NamedValue<V>> entry : this.internalMap().entrySet())
    {
      namedValue = entry.getValue();
      newEntry = new MapEntry<String, V>(namedValue.key(), namedValue.value());
      entries.add(newEntry);
    }
    return entries;
  } // entrySet() 

  // -------------------------------------------------------------------------

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected String standardizeKey(Object key)
  {
    if (key instanceof String)
    {
      String string = (String)key;
      return string.toLowerCase();
    }
    return null;
  } // standardizeKey() 

  // -------------------------------------------------------------------------

  protected Map.Entry<String, NamedValue<V>> findByValue(Object value)
  {
    NamedValue<V> namedValue;

    if (value == null)
    {
      return null;
    }

    for (Map.Entry<String, NamedValue<V>> entry : this.internalMap().entrySet())
    {
      namedValue = entry.getValue();
      if (value.equals(namedValue.value()))
      {
        return entry;
      }
    }
    return null;
  } // findByValue() 

  // -------------------------------------------------------------------------

  protected Map<String, NamedValue<V>> createInternalMap(int capacity)
  {
    return new HashMap<String, NamedValue<V>>(capacity);
  } // createInternalMap() 

  // -------------------------------------------------------------------------

  protected Map<String, NamedValue<V>> internalMap()
  {
    return internalMap;
  } // internalMap() 

  // -------------------------------------------------------------------------

  protected void internalMap(Map<String, NamedValue<V>> newValue)
  {
    internalMap = newValue;
  } // internalMap() 

  // -------------------------------------------------------------------------

} // class CaseInsensitiveKeyMap 
