// ===========================================================================
// CONTENT  : CLASS DateUtil
// AUTHOR   : Manfred Duchrow
// VERSION  : 2.1 - 19/08/2014
// HISTORY  :
//  11/06/2005  mdu  CREATED
//	27/03/2010	mdu		added	-->	constants for time zones and various methods for time zone based date creation
//  19/08/2014  mdu   added --> RFC-1123 support
//
// Copyright (c) 2005-2014, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pf.util ;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Locale;
import java.util.TimeZone;

/**
 * This class is intended to provide convenience methods around the java.util.Date 
 * class.
 * It is not always necessary to fiddle around with the awkward GregorianCalendar
 * and not everybody wants to use the deprecated methods of java.util.Date.
 * <p>
 * <strong>Be aware that this class always treats months 1-based NOT zero-based.
 * That is: 1=January, 2=February, 3=March, ..., 12=December</strong>
 * <p>
 * All date representations of this class are by default based on Locale.getDefault(),
 * but this can be changed via {@link #setDefaultLocale(Locale)}.
 *
 * @author Manfred Duchrow
 * @version 2.1
 */
public class DateUtil
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  public static final TimeZone TIMEZONE_UTC     = TimeZone.getTimeZone("UTC");
  public static final TimeZone TIMEZONE_GMT     = TimeZone.getTimeZone("GMT");
  public static final TimeZone TIMEZONE_GERMANY = TimeZone.getTimeZone("Europe/Berlin");
  
  public static final String SORT_DATE_FORMAT     = "yyyy-MM-dd";
  public static final String GERMAN_DATE_FORMAT   = "dd.MM.yyyy";
  public static final String UK_DATE_FORMAT       = "dd/MM/yyyy";
  public static final String US_DATE_FORMAT       = "MM/dd/yyyy";

  public static final SimpleDateFormat DF_GERMAN_DATE   = new SimpleDateFormat(GERMAN_DATE_FORMAT);
  public static final SimpleDateFormat DF_UK_DATE       = new SimpleDateFormat(UK_DATE_FORMAT);
  public static final SimpleDateFormat DF_US_DATE       = new SimpleDateFormat(US_DATE_FORMAT);
  public static final SimpleDateFormat DF_GERMAN_DATE_TIME   = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss");
  
  /**
   * The date format according to RFC-1123.
   * Example: "Sun, 06 Nov 1994 08:49:37 GMT"
   */
  public static final String RFC_1123_DATE_FORMAT = "EEE, dd MMM yyyy HH:mm:ss z";
  
  private static final String ZULU_DATE_PATTERN = "yyyy-MM-dd'T'HH:mm:ss";

  // =========================================================================
  // CLASS VARIABLES
  // =========================================================================
  private static DateUtil soleInstance = new DateUtil() ;

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private Locale defaultLocale = Locale.getDefault();

  // =========================================================================
  // CLASS METHODS
  // =========================================================================

  /**
   * Returns the only instance this class supports (design pattern "Singleton")
   */
  public static DateUtil current()
  {
    return soleInstance ;
  } // current() 

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  protected DateUtil()
  {
    super() ;
  } // DateUtil() 

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Returns a Date object representing the given parameters.
   * 
   * @param year The year of the date to create 
   * @param month The month of the date to create (1-based, 1=January)
   * @param day The day of the date to create 
   */
  public Date newDate(int year, int month, int day)
  {
    Calendar cal;

    cal = this.newCalendar(year, month, day, 0, 0, 0);
    return cal.getTime();
  } // newDate() 

  // -------------------------------------------------------------------------

  /**
   * Returns a Date object representing the given parameters.
   * 
   * @param timezone The time zone to which the other parameters are related
   * @param year The year of the date to create 
   * @param month The month of the date to create (1-based, 1=January)
   * @param day The day of the date to create 
   */
  public Date newDate(TimeZone timezone, int year, int month, int day)
  {
    Calendar calendar;

    calendar = this.newCalendar(timezone, year, month, day, 0, 0, 0);
    return calendar.getTime();
  } // newDate() 

  // -------------------------------------------------------------------------

  /**
   * Returns a Date object representing the given parameters.
   * 
   * @param year The year of the date to create 
   * @param month The month of the date to create (1-based, 1=January)
   * @param day The day of the date to create 
   * @param hour The hours (0-23)
   * @param minute The minutes (0-59)
   */
  public Date newDate(int year, int month, int day, int hour, int minute)
  {
    Calendar cal;

    cal = this.newCalendar(year, month, day, hour, minute, 0);
    return cal.getTime();
  } // newDate() 

  // -------------------------------------------------------------------------

  /**
   * Returns a Date object representing the given parameters.
   * 
   * @param year The year of the date to create 
   * @param month The month of the date to create (1-based, 1=January)
   * @param day The day of the date to create 
   * @param hour The hours (0-23)
   * @param minute The minutes (0-59)
   * @param second The seconds (0-59)
   */
  public Date newDate(int year, int month, int day, int hour, int minute, int second)
  {
    Calendar cal;

    cal = this.newCalendar(year, month, day, hour, minute, second);
    return cal.getTime();
  } // newDate() 

  // -------------------------------------------------------------------------

  /**
   * Returns a Date object representing the given parameters.
   * 
   * @param year The year of the date to create 
   * @param month The month of the date to create (1-based, 1=January)
   * @param day The day of the date to create 
   * @param hour The hours (0-23)
   * @param minute The minutes (0-59)
   * @param second The seconds (0-59)
   */
  public Calendar newCalendar(int year, int month, int day, int hour, int minute, int second)
  {
    GregorianCalendar cal;
    
    cal = new GregorianCalendar(this.getDefaultLocale());
    cal.set(year, month-1, day, hour, minute, second);
    return cal;
  } // newCalendar() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns a Date object in time zone GMT representing the given parameters.
   * 
   * @param year The year of the date to create 
   * @param month The month of the date to create (1-based, 1=January)
   * @param day The day of the date to create 
   * @param hour The hours (0-23)
   * @param minute The minutes (0-59)
   * @param second The seconds (0-59)
   */
  public Date newGMTDate(int year, int month, int day, int hour, int minute, int second)
  {
    return this.newDate(TIMEZONE_GMT, year, month, day, hour, minute, second);
  } // newGMTDate() 

  // -------------------------------------------------------------------------

  /**
   * Returns a Date object representing the given parameters.
   * 
   * @param timezone The time zone to which the other parameters are related
   * @param year The year of the date to create 
   * @param month The month of the date to create (1-based, 1=January)
   * @param day The day of the date to create 
   * @param hour The hours (0-23)
   * @param minute The minutes (0-59)
   * @param second The seconds (0-59)
   */
  public Date newDate(TimeZone timezone, int year, int month, int day, int hour, int minute, int second)
  {
    Calendar cal;

    cal = this.newCalendar(timezone, year, month, day, hour, minute, second);
    return cal.getTime();
  } // newDate() 

  // -------------------------------------------------------------------------

  /**
   * Returns a Calendar object representing the given parameters.
   * 
   * @param timezone The time zone to which the other parameters are related
   * @param year The year of the date to create 
   * @param month The month of the date to create (1-based, 1=January)
   * @param day The day of the date to create 
   * @param hour The hours (0-23)
   * @param minute The minutes (0-59)
   * @param second The seconds (0-59)
   */
  public Calendar newCalendar(TimeZone timezone, int year, int month, int day, int hour, int minute, int second)
  {
    GregorianCalendar cal;
    
    cal = new GregorianCalendar(timezone, this.getDefaultLocale());
    cal.set(year, month - 1, day, hour, minute, second);
    return cal;
  } // newCalendar() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the current date
   */
  public Date today()
  {
    return new Date();
  } // today() 

  // -------------------------------------------------------------------------

  /**
   * Returns the given date as Calendar in the current machines time zone.
   */
  public Calendar asCalendar(Date date)
  {
    Calendar calendar;

    calendar = new GregorianCalendar(this.getDefaultLocale());
    calendar.setTime(date);
    return calendar;
  } // asCalendar() 

  // -------------------------------------------------------------------------

  /**
   * Returns the given date as Calendar in the specified time zone.
   */
  public Calendar asCalendar(TimeZone timezone, Date date)
  {
    Calendar calendar;

    calendar = new GregorianCalendar(timezone, this.getDefaultLocale());
    calendar.setTime(date);
    return calendar;
  } // asCalendar() 

  // -------------------------------------------------------------------------

  /**
   * Returns the given date as Calendar in the GMT time zone.
   */
  public Calendar asGMTCalendar(Date date)
  {
    return this.asCalendar(TIMEZONE_GMT, date);
  } // asGMTCalendar() 

  // -------------------------------------------------------------------------

  /**
   * Converts the given date into the so-called zulu time.
   * That is: "yyyy-MM-dd'T'HH:mm:ssZ"
   */
  public String convertDateIntoZuluTime(final Date date)
  {
    SimpleDateFormat format = new SimpleDateFormat(ZULU_DATE_PATTERN);
    format.setTimeZone(TIMEZONE_UTC);
    return format.format(date) + "Z";
  } // convertDateIntoZuluTime() 

  // -------------------------------------------------------------------------

  /**
   * Converts the given string that must be formatted as so-called zulu-time
   * to a Date object.
   */
  public Date convertZuluTimeIntoDate(final String str) throws ParseException
  {
    SimpleDateFormat format = new SimpleDateFormat(ZULU_DATE_PATTERN);
    String string = str;
    
    format.setTimeZone(TIMEZONE_UTC);
    if (string.endsWith("Z"))
    {
      string = string.substring(0, string.length() - 1);
    }
    return format.parse(string);
  } // convertZuluTimeIntoDate() 

  // -------------------------------------------------------------------------

  /**
   * Formats the given date according to RFC-1123 which is also used by HTTP/1.1 
   * (see specification section 3.3.1).
   * <p>
   * Example: "Sun, 06 Nov 1994 08:49:37 GMT" ; RFC-1123
   * 
   * @param date The date to format (if null, null will be returned).
   * @return A string representing the given date GMT.
   */
  public String formatRFC1123(Date date) 
  {
    SimpleDateFormat formatter;
    
    if (date == null)
    {
      return null;
    }
    formatter = this.getRFC1123Formatter();
    formatter.setTimeZone(TIMEZONE_GMT);
    return formatter.format(date);
  } // formatRFC1123() 
  
  // -------------------------------------------------------------------------

  /**
   * Parses the given string assuming its format complies with RFC-1123. 
   * 
   * @param value The string to be parsed (if null, null will be returned).
   * @return The date derived from the string.
   * @throws IllegalArgumentException If the string cannot be parsed to a date.
   */
  public Date parseDateRFC1123(String value)
  {
    SimpleDateFormat formatter;

    if (value == null)
    {
      return null;
    }
    formatter = this.getRFC1123Formatter();
    try
    {
      return formatter.parse(value);
    }
    catch (ParseException e)
    {
      throw new IllegalArgumentException("Value " + value + " does not conform to RFC-1123.", e);
    }
  } // parseDateRFC1123() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the default locale used inside all methods of this singleton.
   */
  public Locale getDefaultLocale() 
  {
    return this.defaultLocale;
  } // getDefaultLocale() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Sets the default locale for usage inside all methods of this singleton.
   * If the given locale is null the default will be set to the current
   * platform default locale.
   */
  public void setDefaultLocale(Locale locale) 
  {
    if (locale == null)
    {      
      this.defaultLocale = Locale.getDefault();
    }
    else
    {
      this.defaultLocale = locale;
    }    
  } // setDefaultLocale() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the date format for the current default locale of this singleton.
   * @see #getDefaultLocale()
   * @see #setDefaultLocale(Locale) 
   */
  public String getDefaultDateFormat() 
  {
    return this.getDateFormatForLocale(this.getDefaultLocale());
  } // getDefaultDateFormat() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the date format (only dd MM yyyy) for the given locale.
   * Currently only UK, US, and GERMAN are supported.
   * For all other locales the "yyyy-MM-dd" will be returned. 
   */
  public String getDateFormatForLocale(Locale locale) 
  {
    if (Locale.US.equals(locale))
    {
      return US_DATE_FORMAT;
    }
    if (Locale.UK.equals(locale))
    {
      return UK_DATE_FORMAT;
    }
    if (Locale.GERMAN.equals(locale))
    {
      return GERMAN_DATE_FORMAT;
    }
    return SORT_DATE_FORMAT;
  } // getDateFormatForLocale() 
  
  // -------------------------------------------------------------------------
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected SimpleDateFormat getRFC1123Formatter() 
  {
    // Must always use Locale.US
    return new SimpleDateFormat(RFC_1123_DATE_FORMAT, Locale.US);
  } // getRFC1123Formatter() 
  
  // -------------------------------------------------------------------------
  
} // class DateUtil 
