// ===========================================================================
// CONTENT  : CLASS OrderedSet
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.3 - 18/10/2014
// HISTORY  :
//  12/03/2004  mdu  CREATED
//	19/02/2006	mdu		bugfix	->	removeAll()
//	22/02/2006	mdu		added		-> 	implements List
//  18/10/2014  mdu   changed ->  generic element type
//
// Copyright (c) 2004-2014, by Manfred Duchrow. All rights reserved.
// ===========================================================================
package org.pf.util ;

// ===========================================================================
// IMPORTS
// ===========================================================================
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;

/**
 * Implements a Set that keeps the order of its elements according to when
 * they were added. Additionally it implements the List interface.
 * So instances of this class can either be used as Set or as List.
 *
 * @author Manfred Duchrow
 * @version 1.3
 */
public class OrderedSet<E> implements Set<E>, List<E>
{
	// =========================================================================
	// CONSTANTS
	// =========================================================================

	// =========================================================================
	// INSTANCE VARIABLES
	// =========================================================================
	private List<E> elements = null ;
	protected List<E> getElements() { return elements ; }
	protected void setElements( List<E> newValue ) { elements = newValue ; }	

	// =========================================================================
	// CLASS METHODS
	// =========================================================================

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  /**
   * Initialize the new instance with default values.
   */
  public OrderedSet()
  {
    this(20);
  } // OrderedSet() 

  // -------------------------------------------------------------------------

  /**
   * Initialize the new set with an initial capacity.
   */
  public OrderedSet(int initialCapacity)
  {
    super();
    this.setElements(this.createInnerList(initialCapacity));
  } // OrderedSet() 

  // -------------------------------------------------------------------------  

  /**
   * Initialize the new set with an inner list.
   */
  protected OrderedSet(List<E> innerList)
  {
    super();
    this.setElements(innerList);
  } // OrderedSet() 

  // -------------------------------------------------------------------------  

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * @see java.util.Collection#add(java.lang.Object)
   */
  @Override
  public boolean add(E obj)
  {
    if (!this.getElements().contains(obj))
    {
      return this.getElements().add(obj);
    }
    return false;
  } // add() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#addAll(java.util.Collection)
   */
  @Override
  public boolean addAll(Collection<? extends E> coll)
  {
    boolean changed = false;
    Iterator<? extends E> iter;

    iter = coll.iterator();
    while (iter.hasNext())
    {
      if (this.add(iter.next()))
      {
        changed = true;
      }
    }
    return changed;
  } // addAll() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#clear()
   */
  @Override
  public void clear()
  {
    this.getElements().clear();
  } // clear() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#contains(java.lang.Object)
   */
  @Override
  public boolean contains(Object obj)
  {
    return this.getElements().contains(obj);
  } // contains() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#containsAll(java.util.Collection)
   */
  @Override
  public boolean containsAll(Collection<?> coll)
  {
    return this.getElements().containsAll(coll);
  } // containsAll() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#isEmpty()
   */
  @Override
  public boolean isEmpty()
  {
    return this.getElements().isEmpty();
  } // isEmpty() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#iterator()
   */
  @Override
  public Iterator<E> iterator()
  {
    return this.getElements().iterator();
  } // iterator() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#remove(java.lang.Object)
   */
  @Override
  public boolean remove(Object obj)
  {
    return this.getElements().remove(obj);
  } // remove() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#removeAll(java.util.Collection)
   */
  @Override
  public boolean removeAll(Collection<?> coll)
  {
    return this.getElements().removeAll(coll);
  } // removeAll() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#retainAll(java.util.Collection)
   */
  @Override
  public boolean retainAll(Collection<?> coll)
  {
    return this.getElements().retainAll(coll);
  } // retainAll() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#size()
   */
  @Override
  public int size()
  {
    return this.getElements().size();
  } // size() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#toArray()
   */
  @Override
  public Object[] toArray()
  {
    return this.getElements().toArray();
  } // toArray() 

  // -------------------------------------------------------------------------

  /**
   * @see java.util.Collection#toArray(java.lang.Object[])
   */
  @SuppressWarnings("unchecked")
  @Override
  public Object[] toArray(Object[] array)
  {
    return this.getElements().toArray(array);
  } // toArray() 

  // -------------------------------------------------------------------------

  /**
   * Inserts the specified element at the specified position in this list  
   */
  @Override
  public void add(int index, E element)
  {
    if (!this.contains(element))
    {
      this.getElements().add(index, element);
    }
  } // add() 

  // -------------------------------------------------------------------------

  /**
   *  Inserts all of the elements in the specified collection into this list 
   *  at the specified position if they are not yet in this list.
   *  
   *  @return true if this list changed as a result of the call.
   */
  @Override
  public boolean addAll(int index, Collection<? extends E> c)
  {
    Iterator<? extends E> iter;
    boolean changed = false;
    int insertIndex = index;
    E object;

    iter = c.iterator();
    while (iter.hasNext())
    {
      object = iter.next();
      if (!this.contains(object))
      {
        this.add(insertIndex, object);
        insertIndex++;
        changed = true;
      }
    }
    return changed;
  } // addAll() 

  // -------------------------------------------------------------------------

  /**
   * Returns the element at the specified position in this list.
   */
  @Override
  public E get(int index)
  {
    return this.getElements().get(index);
  } // get() 

  // -------------------------------------------------------------------------

  /**
   * Returns the index in this list of the first occurrence of the specified 
   * element, or -1 if this list does not contain this element.
   */
  @Override
  public int indexOf(Object object)
  {
    return this.getElements().indexOf(object);
  } // indexOf() 

  // -------------------------------------------------------------------------

  /**
   * Returns the index in this list of the last occurrence of the specified 
   * element, or -1 if this list does not contain this element.
   */
  @Override
  public int lastIndexOf(Object object)
  {
    return this.getElements().lastIndexOf(object);
  } // lastIndexOf() 

  // -------------------------------------------------------------------------

  /**
   * Returns a list iterator of the elements in this list (in proper sequence).
   */
  @Override
  public ListIterator<E> listIterator()
  {
    return this.getElements().listIterator();
  } // listIterator() 

  // -------------------------------------------------------------------------

  /**
   * Returns a list iterator of the elements in this list (in proper sequence), 
   * starting at the specified position in this list.
   */
  @Override
  public ListIterator<E> listIterator(int index)
  {
    return this.getElements().listIterator(index);
  } // listIterator() 

  // -------------------------------------------------------------------------

  /**
   * Returns the object at the specified index
   */
  @Override
  public E remove(int index)
  {
    return this.getElements().remove(index);
  } // remove() 

  // -------------------------------------------------------------------------

  /**
   * Sets the given element at the specified index if the element is not
   * yet already in the collection!
   */
  @Override
  public E set(int index, E element)
  {
    if (!this.contains(element))
    {
      return this.getElements().set(index, element);
    }
    return this.get(index);
  } // set() 

  // -------------------------------------------------------------------------

  /**
   * Returns a sub list containing the elements in the range of the defined
   * indices. The result is also an OrderedSet.
   * Contrary to the definition of this method in the List interface here
   * the result list is NOT backed by this list.
   * That means, any changes to the result list do not have any impact on
   * this list!
   */
  @Override
  public List<E> subList(int fromIndex, int toIndex)
  {
    List<E> result;

    result = this.createInnerList(toIndex - fromIndex);
    for (int i = fromIndex; i < toIndex; i++)
    {
      result.add(this.get(i));
    }
    return new OrderedSet<E>(result);
  } // subList() 

  // -------------------------------------------------------------------------

  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected List<E> createInnerList(int initialCapacity)
  {
    return new ArrayList<E>(initialCapacity);
  } // createInnerList() 

  // -------------------------------------------------------------------------

} // class OrderedSet 
