// ===========================================================================
// CONTENT  : CLASS SimpleJsonParser
// AUTHOR   : Manfred Duchrow
// VERSION  : 1.0 - 24/08/2014
// HISTORY  :
//  24/08/2014  mdu  CREATED
//
// Copyright (c) 2014, by MDCS. All rights reserved.
// ===========================================================================
package org.pfsw.text.json;

// ===========================================================================
// IMPORTS
// ===========================================================================
import static org.pfsw.text.json.JsonConstants.*;

import java.math.BigDecimal;
import java.util.Map;

import org.pf.text.StringExaminer;
import org.pf.text.StringUtil;

/**
 * A parser that creates a Java internal structure of JsonObject and
 * JsonArray objects from a given JSON text.
 * <p>
 * The data will be mapped on the following Java types:
 * <dl>
 * <dt>Object</dt><dd>JsonObject</dd>
 * <dt>Array</dt><dd>JsonArray</dd>
 * <dt>null</dt><dd>null</dd>
 * <dt>true</dt><dd>Boolean.TRUE</dd>
 * <dt>false</dt><dd>Boolean.FALSE</dd>
 * <dt>"string"</dt><dd>String</dd>
 * <dt>integer number</dt><dd>Integer</dd>
 * <dt>long number</dt><dd>Long</dd>
 * <dt>other numbers</dt><dd>BigDecimal</dd>
 * </dl>
 * <p>
 * It is based on the RFC-7159 (http://tools.ietf.org/html/rfc7159).
 *
 * @author Manfred Duchrow
 * @version 1.0
 */
public class SimpleJsonParser
{
  // =========================================================================
  // CONSTANTS
  // =========================================================================
  private static final StringUtil SU = StringUtil.current();

  // =========================================================================
  // INSTANCE VARIABLES
  // =========================================================================
  private final CharEscaper charEscaper = new CharEscaper();
  private StringExaminer stringExaminer;

  // =========================================================================
  // CONSTRUCTORS
  // =========================================================================
  public SimpleJsonParser()
  {
    super();
  } // SimpleJsonParser() 

  // =========================================================================
  // PUBLIC INSTANCE METHODS
  // =========================================================================
  /**
   * Parse the given JSON text and return its Java representation. 
   * That is one of the following types: 
   * <ul>
   * <li>JsonObject</li>
   * <li>JasonArray</li>
   * <li>String</li>
   * <li>Boolean</li>
   * <li>Integer</li>
   * <li>Long</li>
   * <li>BigDecimal</li>
   * <li>null</li>
   * </ul> 
   * 
   * @param jsonText The JSON text to be parsed (must not be null).
   * @return The parsed data as Java objects.
   * @throws JsonParseException Any parsing problem.
   */
  public Object parse(final String jsonText) throws JsonParseException
  {
    this.setStringExaminer(new StringExaminer(jsonText));
    return this.parseInternal();
  } // parse() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Parse the given JSON text and return its Java representation as JsonObject. 
   * That implies that the given JSON text starts with a "{".
   * 
   * @param jsonText The JSON text to be parsed (must not be null).
   * @return The parsed data as Java objects.
   * @throws JsonParseException Any parsing problem.
   * @throws ClassCastException If the JSON text did not represent a JSON object.
   */
  public JsonObject parseObject(String jsonText) throws JsonParseException 
  {
    return this.parseTo(jsonText, JsonObject.class);
  } // parseObject() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Parse the given JSON text and return its Java representation as JsonArray. 
   * That implies that the given JSON text starts with a "[".
   * 
   * @param jsonText The JSON text to be parsed (must not be null).
   * @return The parsed data as Java objects.
   * @throws JsonParseException Any parsing problem.
   * @throws ClassCastException If the JSON text did not represent a JSON array.
   */
  public JsonArray parseArray(String jsonText) throws JsonParseException 
  {
    return this.parseTo(jsonText, JsonArray.class);
  } // parseArray() 
  
  // -------------------------------------------------------------------------
  
  // =========================================================================
  // PROTECTED INSTANCE METHODS
  // =========================================================================
  protected boolean isObjectStart(final char ch)
  {
    return (BEGIN_OBJECT == ch);
  } // isObjectStart() 

  // -------------------------------------------------------------------------
  
  protected boolean isObjectEnd(final char ch)
  {
    return (END_OBJECT == ch);
  } // isObjectEnd() 
  
  // -------------------------------------------------------------------------
  
  protected boolean isArrayStart(final char ch)
  {
    return (BEGIN_ARRAY == ch);
  } // isArrayStart() 

  // -------------------------------------------------------------------------
  
  protected boolean isArrayEnd(final char ch)
  {
    return (END_ARRAY == ch);
  } // isArrayEnd() 
  
  // -------------------------------------------------------------------------
  
  protected boolean isStringStart(final char ch)
  {
    return (this.getStringDelimiter() == ch);
  } // isStringStart() 
  
  // -------------------------------------------------------------------------
  
  protected boolean isStringEnd(final char ch)
  {
    return (this.getStringDelimiter() == ch);
  } // isStringEnd() 
  
  // -------------------------------------------------------------------------
  
  protected boolean isEscape(final char ch)
  {
    return (ESCAPE == ch);
  } // isEscape() 
  
  // -------------------------------------------------------------------------
  
  protected boolean isValueSeparator(final char ch)
  {
    return (VALUE_SEPARATOR == ch);
  } // isValueSeparator() 
  
  // -------------------------------------------------------------------------
  
  protected boolean isNumberStart(final char ch)
  {
    return Character.isDigit(ch) || (ch == '-') || (ch == '+');
  } // isNumberStart() 
  
  // -------------------------------------------------------------------------
  
  protected Object parseInternal() throws JsonParseException
  {
    // According to RFC-7159 all values are allowed, not only Object and Array
    return this.parseValue();
  } // parseInternal() 
  
  // -------------------------------------------------------------------------
  
  protected Object parseValue() throws JsonParseException
  {
    char ch;
    
    ch = this.peekNextNoneWhitespaceChar();
    
    if (this.isObjectStart(ch))
    {
      return this.parseObject();
    }
    if (this.isArrayStart(ch))
    {
      return this.parseArray();
    }
    if (this.isStringStart(ch))
    {
      return this.parseString();
    }
    if (this.isNumberStart(ch))
    {
      return this.parseNumber();
    }
    return this.parseLiteral();
  } // parseValue() 
  
  // -------------------------------------------------------------------------
  
  protected JsonObject parseObject() throws JsonParseException
  {
    JsonObject jsonObject;
    char ch;
    
    this.checkNextMandatoryToken(BEGIN_OBJECT);
    jsonObject = this.createObject();

    ch = this.peekNextNoneWhitespaceChar();
    if (this.isObjectEnd(ch))
    {
      this.nextChar(); // Read it away
      return jsonObject;
    }
    
    while (true)
    {      
      this.parseKeyValueInto(jsonObject);
      
      ch = this.nextNoneWhitespaceChar();
      if (this.isObjectEnd(ch))
      {
        return jsonObject;
      }
      if (!this.isValueSeparator(ch))
      {
        throw new JsonParseException("Value separator '" + VALUE_SEPARATOR + "' expected at position: " + this.getLastPosition());
      }
    }
  } // parseObject() 
  
  // -------------------------------------------------------------------------

  protected void parseKeyValueInto(final Map<String, Object> jsonObject) throws JsonParseException
  {
    String key;
    Object value;

    key = this.parseString();
    this.checkNextMandatoryToken(NAME_SEPARATOR);
    value = this.parseValue();
    jsonObject.put(key, value);
  } // parseKeyValueInto() 
  
  // -------------------------------------------------------------------------
  
  protected JsonArray parseArray() throws JsonParseException
  {
    JsonArray jsonArray;
    char ch;
    Object value;
    
    this.checkNextMandatoryToken(BEGIN_ARRAY);
    jsonArray = this.createArray();
    
    ch = this.peekNextNoneWhitespaceChar();
    if (this.isArrayEnd(ch))
    {
      this.nextChar(); // Read it away
      return jsonArray;
    }
    
    while (true)
    {      
      value = this.parseValue();
      jsonArray.add(value);
      
      ch = this.nextNoneWhitespaceChar();
      if (this.isArrayEnd(ch))
      {
        return jsonArray;
      }
      if (!this.isValueSeparator(ch))
      {
        throw new JsonParseException("Value separator '" + VALUE_SEPARATOR + "' expected in array at position: " + this.getLastPosition());
      }
    }
  } // parseArray() 

  // -------------------------------------------------------------------------
  
  protected String parseString() throws JsonParseException
  {
    String string;

    this.checkNextMandatoryToken(this.getStringDelimiter());
    string = this.parseUpToQuote();
    return string;
  } // parseString() 
  
  // -------------------------------------------------------------------------
  
  protected String parseUpToQuote() throws JsonParseException
  {
    StringBuilder buffer;
    char ch;
    
    buffer = new StringBuilder(200);
    ch = this.nextChar();
    while (!this.isStringEnd(ch))
    {
      if (this.isEscape(ch))
      {
        ch = this.nextChar();
        ch = this.getCharEscaper().getCharForEscapedChar(ch);
      }
      buffer.append(ch);
      ch = this.nextChar();
    }
    return buffer.toString();
  } // parseUpToQuote() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns either an Integer, a Long or a BigDecimal.

   * @throws JsonParseException If the string cannot be parsed to a number.
   */
  protected Number parseNumber() throws JsonParseException
  {
    String string;
    BigDecimal number;
    int pos;
    
    pos = this.getLastPosition();
    string = this.parseLiteralString(30, "+-0123456789.eE");
    try
    {
      number = new BigDecimal(string);
    }
    catch (NumberFormatException ex)
    {
      throw new JsonParseException("Invalid number: " + string + " , at position: " + pos);
    }
    if (number.scale() == 0)
    {
      if (number.compareTo(BigDecimal.valueOf(Integer.MAX_VALUE)) < 0)
      {
        return Integer.valueOf(number.intValue());        
      }
      return Long.valueOf(number.longValue());        
    }
    return number;
  } // parseNumber() 
  
  // -------------------------------------------------------------------------
  
  protected String parseLiteralString(int maxLength, String allowedChars) throws JsonParseException
  {
    StringBuilder buffer;
    char ch;
    
    buffer = new StringBuilder(maxLength);
    ch = this.peekNextChar();
    while ((buffer.length() < maxLength) && (SU.contains(allowedChars, ch)))
    {
      buffer.append(this.nextChar());
      ch = this.peekNextChar();
    }
    
    return buffer.toString();
  } // parseLiteralString() 
  
  // -------------------------------------------------------------------------
  
  protected Object parseLiteral() throws JsonParseException
  {
    String text;
    int pos;
    
    this.peekNextNoneWhitespaceChar();
    pos = this.getLastPosition();
    text = this.parseLiteralString(5, "truefalsn");
    if (TRUE.equals(text))
    {
      return Boolean.TRUE;
    }
    if (FALSE.equals(text))
    {
      return Boolean.FALSE;
    }
    if (NULL.equals(text))
    {
      return null;
    }
    throw new JsonParseException("No valid JSON literal: " + text + "   at position " + pos);    
  } // parseLiteral() 
  
  // -------------------------------------------------------------------------
  
  @SuppressWarnings("unchecked")
  protected <T extends JsonType> T parseTo(final String json, final Class<T> resultType) throws JsonParseException
  {
    Object result;
    
    result = this.parse(json);
    if (resultType.isInstance(result))
    {      
      return (T)result;
    }
    throw new JsonParseException("Class cast problem for parsed JSON. Actual class is: " + result.getClass().getName());
  }
  
  // -------------------------------------------------------------------------
  
  protected int getLastPosition()
  {
    return this.getStringExaminer().getPosition();
  } // getLastPosition() 
  
  // -------------------------------------------------------------------------

  protected void checkNextMandatoryToken(char expectedChar) throws JsonParseException
  {
    char ch;
    
    ch = this.nextNoneWhitespaceChar();    
    if (ch != expectedChar)
    {
      throw new JsonParseException("Expected token <" + expectedChar + "> not found at position " 
            + this.getLastPosition());        
    }
  } // checkNextMandatoryToken() 

  // -------------------------------------------------------------------------

  /**
   * Looks for the next character that is no white space and returns it.
   * Position when finished: on next character after the one returned. 
   * @throws JsonParseException If the end of the underlying JSON data was reached.
   */
  protected char nextNoneWhitespaceChar() throws JsonParseException 
  {
    char ch;
    
    ch = this.getStringExaminer().nextNoneWhitespaceChar();
    if (this.getStringExaminer().endReached(ch))
    {
      throw new JsonParseException("End of data reached unexpectedly.");
    }
    return ch;
  } // nextNoneWhitespaceChar() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns simply the next character of the underlying JSON string.
   * Position when finished: on next character after the one returned. 
   * @throws JsonParseException If the end of the underlying JSON data was reached.
   */
  protected char nextChar() throws JsonParseException 
  {
    char ch;
    
    ch = this.getStringExaminer().nextChar();
    if (this.getStringExaminer().endReached(ch))
    {
      throw new JsonParseException("Unexpected end of data reached.");
    }
    return ch;
  } // nextChar() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Returns the next character of the underlying JSON string without moving the 
   * examiner's position forward. That is, the next nextChar() call will return 
   * the same character again.
   * Position when finished: still on same character. 
   */
  protected char peekNextChar() throws JsonParseException 
  {
    return this.getStringExaminer().peek();
  } // peekNextChar() 
  
  // -------------------------------------------------------------------------
  
  /**
   * Looks for the next character that is no white space and returns it.
   * Position when finished: on the character returned. 
   * @throws JsonParseException If the end of the underlying JSON data was reached.
   */
  protected char peekNextNoneWhitespaceChar() throws JsonParseException 
  {
    char ch;

    ch = this.nextNoneWhitespaceChar();
    this.getStringExaminer().skip(-1);
    return ch;
  } // peekNextNoneWhitespaceChar() 
  
  // -------------------------------------------------------------------------
  
  protected char getStringDelimiter() 
  {
    return STRING_DELIMITER;
  } // getStringDelimiter() 
  
  // -------------------------------------------------------------------------
  
  protected JsonObject createObject()
  {
    return new JsonObject();
  } // createObject() 

  // -------------------------------------------------------------------------

  protected JsonArray createArray()
  {
    return new JsonArray();
  } // createArray() 
  
  // -------------------------------------------------------------------------
  
  protected StringExaminer getStringExaminer()
  {
    return stringExaminer;
  } // getStringExaminer() 
  
  // -------------------------------------------------------------------------

  protected void setStringExaminer(StringExaminer stringExaminer)
  {
    this.stringExaminer = stringExaminer;
  } // setStringExaminer() 
  
  // -------------------------------------------------------------------------

  public CharEscaper getCharEscaper()
  {
    return charEscaper;
  } // getCharEscaper()
  
  // -------------------------------------------------------------------------
  
  /**
   * Just for debugging purposes!
   */
  protected String getCurrentStringRegion(int backwards, int forwards) 
  {
    int len;
    StringBuilder buffer;
    
    len = backwards + forwards;
    buffer = new StringBuilder(len);
    this.getStringExaminer().markPosition();
    this.getStringExaminer().skip(backwards * (-1));
    for (int i = 0; i < len; i++)
    {
      buffer.append(this.getStringExaminer().nextChar());
    }
    this.getStringExaminer().restorePosition();
    return buffer.toString();
  } // getCurrentStringRegion() 

  // -------------------------------------------------------------------------
  
} // class SimpleJsonParser 
